
/*
 * netf: network information filter
 *
 * by Oguma Keiichiro
 * From 1996. 8. 19
 */

#include <winsock.h>
#include <stdio.h>
#include <time.h>

#define MAX_HOSTNAME 128
#define MAX_TOKEN 64
#define ESCAPE '%'
#define MAX_BUF 512
#define MAXIP 10

#define STRERROR "!error!"
#define STRILLTOKEN "!token!"

static time_t Now;
static struct tm *Tmst;
static char Hostname[MAX_HOSTNAME] = "";
static unsigned char IpAddress[MAXIP][4];
static enum {
  NET_INIT,
  NET_OK,
  NET_ERROR
} NetStatus = NET_INIT;

static int convert(FILE *fpin, FILE *fpout);
static void output(char *token, FILE *fp);
static char *getNetInfo(void);
static char *errorMessage(int errorCode);
static void usage(int inDetail);


void main(int argc, char **argv)
{
  FILE *fpin = NULL;
  FILE *fpout = NULL;

  time(&Now);
  Tmst = localtime(&Now);
  
  while(--argc > 0) {
    ++argv;
    if ((*argv)[0] == '-' && (*argv)[1] != '\0') {
      ++*argv;
      switch(**argv) {
      case 'h':
        usage(TRUE);
        exit(1);
      default:
        usage(FALSE);
        exit(1);
      }
    }
    else if (fpin == NULL) {
      if (**argv == '-')
        fpin = stdin;
      else if ((fpin = fopen(*argv, "rt")) == NULL) {
        fprintf(stderr, "can't open %s\n", *argv);
        exit(1);
      }
    }
    else if (fpout == NULL) {
      if (**argv == '-')
        fpout = stdout;
      if ((fpout = fopen(*argv, "wt")) == NULL) {
        fprintf(stderr, "can't open %s\n", *argv);
        exit(1);
      }
    }
    else {
      usage(FALSE);
      exit(1);
    }
  }
  if (fpin == NULL)
    fpin = stdin;
  if (fpout == NULL)
    fpout = stdout;
  convert(fpin, fpout);
  exit(0);
}


static int convert(FILE *fpin, FILE *fpout)
{
  int c;
  char token[MAX_TOKEN];
  int tokenLen;
  int inToken = FALSE;

  while((c = fgetc(fpin)) != EOF) {
    if (c == ESCAPE) {
      if (inToken) {
        token[tokenLen] = '\0';
        output(token, fpout);
        inToken = FALSE;
      }
      else {
        inToken = TRUE;
        tokenLen = 0;
      }
    }
    else if (inToken) {
      token[tokenLen++] = c;
      token[tokenLen] = '\0';
      if ((!isalnum(c) && c != '-') || tokenLen == MAX_TOKEN - 1) {
        fprintf(fpout, "%%%s", token);
        inToken = FALSE;
      }
    }
    else
      fputc(c, fpout);
  }
  return(0);
}


static void output(char *token, FILE *fp)
{
  char buf[MAX_BUF];
  int digit = FALSE;
  int toLeft = FALSE;
  int width = 0;
  int zeroPad = FALSE;
  int value;
  char *p = token;

  if (*p == '-') {
    toLeft = TRUE;
    ++p;
  }
  if (*p == '0') {
    zeroPad = TRUE;
    ++p;
  }
  while(isdigit(*p)) {
    width = width * 10 + *p++ - '0';
  }
  if (width > sizeof(buf) - 1)
    width = sizeof(buf) - 1;
  
  if (stricmp(p, "year") == 0) {
    digit = TRUE;
    value = Tmst->tm_year + 1900;
  }
  else if (stricmp(p, "shortyear") == 0) {
    digit = TRUE;
    value = Tmst->tm_year % 100;
  }
  else if (stricmp(p, "month") == 0) {
    digit = TRUE;
    value = Tmst->tm_mon + 1;
  }
  else if (stricmp(p, "mday") == 0) {
    digit = TRUE;
    value = Tmst->tm_mday;
  }
  else if (stricmp(p, "hour") == 0) {
    digit = TRUE;
    value = Tmst->tm_hour;
  }
  else if (stricmp(p, "min") == 0) {
    digit = TRUE;
    value = Tmst->tm_min;
  }
  else if (stricmp(p, "sec") == 0) {
    digit = TRUE;
    value = Tmst->tm_sec;
  }
  else if (stricmp(p, "yday") == 0) {
    digit = TRUE;
    value = Tmst->tm_yday + 1;
  }
  else if (stricmp(p, "wday") == 0) {
    digit = TRUE;
    value = Tmst->tm_wday + 1;
  }
  else if (stricmp(p, "day") == 0) {
    char *days[] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
    
    strcpy(buf, days[Tmst->tm_wday]);
  }
  else if (strnicmp(p, "ip", 2) == 0) {
    int n = 0;
    
    p += 2;
    while(*p != '\0') {
      if (isdigit(*p))
        n = n * 10 + *p - '0';
      else {
        fprintf(fp, "%%%s%%", token);
        return;
      }
      ++p;
    }
    if ((p = getNetInfo()) != NULL) {
      strcpy(buf, p);
    }
    else {
      if (n < 0 || n >= MAXIP || (IpAddress[n][0] | IpAddress[n][1] | IpAddress[n][2] | IpAddress[n][3]) == 0) {
        sprintf(buf, "!Illegal IP number!");
      }
      else
        sprintf(buf, zeroPad ? "%03u.%03u.%03u.%03u" : "%u.%u.%u.%u", IpAddress[n][0], IpAddress[n][1], IpAddress[n][2], IpAddress[n][3]);
    }
  }
  else if (stricmp(p, "hostname") == 0) {
    if ((p = getNetInfo()) != NULL) {
      strcpy(buf, p);
    }
    else
      sprintf(buf, "%s", Hostname);
  }
  else {
    fprintf(fp, "%%%s%%", token);
    return;
  }
  if (digit)
    fprintf(fp, zeroPad ? "%0*d" : "%*d", toLeft ? -width : width, value);
  else
    fprintf(fp, "%*s", toLeft ? -width : width, buf);
}


static char *getNetInfo()
{
  struct hostent *host;
  WORD requestedVersion;
  WSADATA wsadata;

  if (NetStatus == NET_OK)
    return(NULL);
  if (NetStatus == NET_ERROR)
    return(STRERROR);
  requestedVersion = MAKEWORD(1, 1); // needs winsock version 1.1 or higher
  switch(WSAStartup(requestedVersion, &wsadata)) {
  case WSASYSNOTREADY:
    // system not ready to communicate
    return("!system is not ready to communicate!");
  case WSAVERNOTSUPPORTED:
    // winsock service is not provided
    return("!winsock service is not provided!");
  case WSAEINVAL:
    // winsok version is old
    return("!winsock version is old!");
  case 0:
    // OK
    break;
  default:
    // unknown error
    return("!unknown winsock error!");
  }
  
  if (gethostname(Hostname, sizeof(Hostname))) {
    NetStatus = NET_ERROR;
    return(errorMessage(WSAGetLastError()));
  }
  if ((host = gethostbyname(Hostname)) == NULL) {
    NetStatus = NET_ERROR;
    return(errorMessage(WSAGetLastError()));
  }
  if (host->h_length != 4) {
    NetStatus = NET_ERROR;
    return("!Illegal Address!");
  }
  {
    int i;

    for(i = 0; i < MAXIP; ++i) {
      if (host->h_addr_list[i] == NULL) {
        for(; i < MAXIP; ++i) {
          IpAddress[i][0] = 0;
          IpAddress[i][1] = 0;
          IpAddress[i][2] = 0;
          IpAddress[i][3] = 0;
        }
        break;
      }
      IpAddress[i][0] = (unsigned char)host->h_addr_list[i][0];
      IpAddress[i][1] = (unsigned char)host->h_addr_list[i][1];
      IpAddress[i][2] = (unsigned char)host->h_addr_list[i][2];
      IpAddress[i][3] = (unsigned char)host->h_addr_list[i][3];
    }
  }
  NetStatus = NET_OK;
  return(NULL);
}


static char *errorMessage(int errorCode)
{
  switch(errorCode) {
  case WSAEFAULT:
    // hostname size is too small
    return("!hostname size is too small!");
  case WSANOTINITIALISED:
    // not initialized
    return("!winsock is not initialized!");
  case WSAENETDOWN:
    // winsock is down
    return("!winsock is down!");
  case WSAEINPROGRESS:
    // blocking operation is in progress
    return("!blocking operation is in progress!");
  case WSAHOST_NOT_FOUND:
    // host not found
    return("!host not found!");
  case WSATRY_AGAIN:
    // DNS is down
    return("!DNS is down!");
  case WSANO_RECOVERY:
    // unrecoverable error
    return("!unrecoverable error!");
  case WSANO_DATA:
    // no data
    return("!no data!");
  case WSAEINTR:
    // blocking operation is cancelled
    return("!blocking operation is cancelled!");
    break;
  default:
    return("!unknwon error!");
  }
}


static void usage(int inDetail)
{
  fputs(
    "netf Version 1.0\n"
  , stderr);
  if (!inDetail) {
    fputs(
      "netf Version 1.0\n"
      "usage: netf [<infile> [<outfile>]]\n"
      "  netf converts some tokens in the data in the file or from the pipe.\n"
      "  netf -h for more details.\n"
    , stderr);
  }
  else {
    fputs(
      "usage: netf [<infile> [<outfile>]]\n"
      "  prefixes\n"
      "   -     to left                            1-9   width\n"
      "   0     zero padding (only for digits)\n"
      "  tokens\n"
      "   %year%      year (4 digits)              %shortyear% year (2 digits)\n"
      "   %month%     month (1 - 12)               %mday%      day of month (1-31)\n"
      "   %hour%      hour (0-23)                  %min%       minute (0-59)\n"
      "   %sec%       second (0-59)                %yday%      day of year (1-366)\n"
      "   %wday%      day of week (1-7, 1:Sunday)  %day%       day (Sun-Sat)\n"
      "   %hostname%  host name                    %ip%        IP address (?.?.?.?)\n"
      "  sample\n"
      "   %-10day%    'Wed       '                 %05hour%    '00011\n"
    , stderr);
  }
  fputs(
    "\n"
    "  Produced 1996,1997 by 'bak' Oguma."
  , stderr);
}


